$(document).ready(function () {
  const wordsearchContainers = $('.wordsearch')
  if (wordsearchContainers.length > 0) {
    for (let i = 0; i < wordsearchContainers.length; i++) {
      const gameDataId = $(wordsearchContainers[i]).attr('data-game-data-var');
      wordSearchInit(wordsearchContainers[i], gameDataId);
    }
  }
});

const CELL_SPACING = 5; // used for "touched/dragged" element calculation
const TABLE_PADDING = 8;

function generatePrompt() {
  const container = $('<div class="prompt">');
  const content = $('<p>', {
    html: 'Sve riječi su uspješno pronađene. Bravo!',
  });

  content.appendTo(container);

  return container;
};

function resetWordSearch(gameContainer, gameId) {
  const oldLetters = $(gameContainer).children('.letters');
  oldLetters.remove();

  wordSearchInit(gameContainer, gameId);
}

function wordSearchInit(gameContainer, gameId) {
  const gameData = window[gameId];

  let gameOver = false;

  const words = gameData[0].solution.map(function (solution) { return solution.word });
  const gameInfo = gameData[0].rows ? { grid: gameData[0].rows, unplaced: [] } : generateGrid(words, gameData[0].options || {});

  if ($(gameContainer).children('.prompt').length > 0) $(".prompt").remove();
  const endGamePrompt = generatePrompt();

  let wordsToSolve = words.filter(function (word) {
    if (gameInfo.unplaced.length) {
      return !gameInfo.unplaced.includes(word);
    } else {
      return word;
    }
  });

  const letters = generateLetters(gameInfo.grid);
  const hints = generateHints(gameData[0].solution);

  let startingLetter;
  let secondLetter;
  let dirX;
  let dirY;
  let currentSelection = [];

  if ($(gameContainer).children('.resetButton').length === 0) {
    const resetDiv = $('<div class="resetButton">');
    const reset = $('<button>', {
      html: gameData[0].options && gameData[0].options.resetButton || 'Ponovi',
      class: 'button'  
    });
    reset.click(function () {
      resetWordSearch(gameContainer, gameId);
    });

    reset.appendTo(resetDiv);
    resetDiv.appendTo(gameContainer);
  }

  if (letters) {
    letters.appendTo(gameContainer);
    /*endGamePrompt.appendTo(letters); BT*/
	endGamePrompt.appendTo(gameContainer);
  } else {
    console.error('Letters not found');
  }

  if (hints && hints.length > 0) {
    const hintsContainer = $('<div class="hints">');
    for (let i = 0; i < hints.length; i++) {
      hints[i].appendTo(hintsContainer);
    }
    hintsContainer.appendTo(gameContainer);
  }

  function generateHints(data) { // creates and returns array of hintDivs, returns empty array if there is none
    let hintElements = [];
    for (let i = 0; i < data.length; i++) {
      const hintDiv = $('<div>', {
        hintId: data[i].id
      });

      let hintContent;

      switch (data[i].hint && data[i].hint.type) {
        case 'self': {
          hintContent = $('<span>', {
            html: data[i].word,
          });
          break;
        }
        case 'text': {
          hintContent = $('<span>', {
            html: data[i].hint.content,
          });
          break;
        }
        case 'img': {
          hintContent = $('<img>', {
            src: data[i].hint.content,
          });
          break;
        }
        case 'nohint': {
          break;
        }
        default: {
          break;
        }
      }

      if (hintContent) {
        hintContent.appendTo(hintDiv);
        hintElements.push(hintDiv);
      }
    };

    return hintElements;
  }

  function generateLetters(data) { // creates <tr> and <td> elements, attaches event listeners, and returns them in array
    const lettersContainer = $('<div class="letters">');
    const table = $('<table>');
    for (let i = 0; i < data.length; i++) {
      const tr = $('<tr>');
      const letters = data[i].split('');

      for (let j = 0; j < letters.length; j++) {
        const letter = letters[j];
        const td = $('<td>', {
          value: letter,
          html: letter,
          x: j,
          y: i,
        });

        $(td).on({
          pointerdown: function (e) {
            handleSelection(td);
          },
          pointermove: function (e) {
            if (e.buttons > 0) {
              handlePointerMove(table);
            }
          },
          touchmove: function (e) {
            handleTouchMove(table);
          }
        });

        td.appendTo(tr);
      }
      tr.appendTo(table);
    }
    table.appendTo(lettersContainer);

    return lettersContainer;
  }

  function setDirection(letter) { // locks the direction of selection after the second letter has been selected
    if (secondLetter) {
      for (let i = 1; i < currentSelection.length; i++) {
        currentSelection[i].classList.remove('selected');
      }
      currentSelection = [];
      currentSelection.push(startingLetter);
    }
    secondLetter = letter;
    dirX = letter.getAttribute('x') - startingLetter.getAttribute('x');
    dirY = letter.getAttribute('y') - startingLetter.getAttribute('y');
  }

  function isMoveLegal(nextLetter) { // checks if a move can be made
    if (!nextLetter) {
      return false;
    }
    if (currentSelection.length < 2) {
      return true;
    }
    const lastLetter = currentSelection[currentSelection.length - 1];
    let xIsValid;
    let yIsValid;

    xIsValid = nextLetter.getAttribute('x') - lastLetter.getAttribute('x') === dirX;
    yIsValid = nextLetter.getAttribute('y') - lastLetter.getAttribute('y') === dirY;

    return xIsValid && yIsValid;
  }

  function isAdjacentToFirstLetter(nextLetter) {
    if (nextLetter && startingLetter) {
      xIsAdjacent = Math.abs(nextLetter.getAttribute('x') - startingLetter.getAttribute('x')) === 1 || nextLetter.getAttribute('x') - startingLetter.getAttribute('x') === 0;
      yIsAdjacent = Math.abs(nextLetter.getAttribute('y') - startingLetter.getAttribute('y')) === 1 || nextLetter.getAttribute('y') - startingLetter.getAttribute('y') === 0;

      return xIsAdjacent && yIsAdjacent;
    }
  }

  function checkWord(word) {                  // takes in constructed string, checks if it matches any string in the solutions (or the inverted version)
    const invertedWord = invertString(word);  // if it finds a match it "solves" it

    const isMatch = gameData[0].solution.filter(function (el) {
      return el.word.toLowerCase() == word.toLowerCase() || el.word.toLowerCase() == invertedWord.toLowerCase();
    })[0];

    if (isMatch) {
      for (let i = 0; i < currentSelection.length; i++) {
        if (!currentSelection[i].classList.contains('solved')) {
          currentSelection[i].classList.add('solved')
        }
      }

      for (let i = 0; i < hints.length; i++) {
        const hintInSolutions = $($(hints[i]).children()[0]);

        if (hints[i].attr('hintId') == isMatch.id && !hints[i][0].classList.contains('solved')) {
          if (hintInSolutions.text() && hintInSolutions.text() != isMatch.word) { // if solution word is already used as a hint
            hintInSolutions.text(isMatch.word.toUpperCase());
          }
          hints[i][0].classList.add('solved');
        }
      }

      solveWord(word, invertedWord);
      resetMove();
    }
  }

  function solveWord(word, invertedWord) { // removes the word from array of words to solve so we can trigger "end game"
    let indexToRemove = undefined;
    for (let i = 0; i < wordsToSolve.length; i++) {
      if (wordsToSolve[i].toLowerCase() === word.toLowerCase() || wordsToSolve[i].toLowerCase() === invertedWord.toLowerCase()) {
        indexToRemove = i;
      }
    }

    if (indexToRemove || indexToRemove === 0) {
      wordsToSolve.splice(indexToRemove, 1);
    }

    if (!wordsToSolve.length) {
      endGame();
    }
  }

  function endGame() {
    $(endGamePrompt).css('display', 'flex');
    gameOver = true;
  }

  function resetMove() {
    for (let i = 0; i < currentSelection.length; i++) {
      currentSelection[i].classList.remove('selected');
    }
    currentSelection = [];
    startingLetter = null;
    dirX = null;
    dirY = null;
  }

  function addToSelection(letter) {
    currentSelection.push(letter);
    letter.classList.add('selected');
  }

  function removeFromSelection(letter) {
    let indexToRemove = undefined;
    for (let i = 0; i < currentSelection.length; i++) {
      if (currentSelection[i] === letter) {
        indexToRemove = i;
      }
    }

    if (indexToRemove !== currentSelection.length - 1 || indexToRemove === 0) { // if selected any other than the last, keep it in selection
      indexToRemove += 1;
    }

    for (let i = indexToRemove; i < currentSelection.length; i++) {
      currentSelection[i].classList.remove('selected');
    }

    currentSelection.splice(indexToRemove, currentSelection.length);
  }

  // EVENTS

  function handleTouchMove(table) {
    event.preventDefault();
    const lastLetter = currentSelection[currentSelection.length - 1];
    const x = Math.floor(((event.pageX - $(table).offset().left) - TABLE_PADDING) / ($('.wordsearch').find('td')[0].clientWidth + CELL_SPACING));
    const y = Math.floor(((event.pageY - $(table).offset().top) - TABLE_PADDING) / ($('.wordsearch').find('td')[0].clientHeight + CELL_SPACING));

    const touchedElement = table.children().find('td').toArray().filter(function (el) {
      return Number($(el).attr('x')) === x && Number($(el).attr('y')) === y;
    });

    if (touchedElement[0] !== lastLetter) {
      if (isAdjacentToFirstLetter(touchedElement[0])) {
        setDirection(touchedElement[0]);
      }
      if (isMoveLegal(touchedElement[0])) {
        handleSelection(touchedElement);
      }
    }
  }

  function handlePointerMove(table) {
    const lastLetter = currentSelection[currentSelection.length - 1];
    const x = Math.floor(((event.pageX - $(table).offset().left) - TABLE_PADDING) / ($('.wordsearch').find('td')[0].clientWidth + CELL_SPACING));
    const y = Math.floor(((event.pageY - $(table).offset().top) - TABLE_PADDING) / ($('.wordsearch').find('td')[0].clientHeight + CELL_SPACING));

    const touchedElement = table.children().find('td').toArray().filter(function (el) {
      return Number($(el).attr('x')) === x && Number($(el).attr('y')) === y;
    });

    if (touchedElement[0] !== lastLetter) {
      if (isAdjacentToFirstLetter(touchedElement[0])) {
        setDirection(touchedElement[0]);
      }
      if (isMoveLegal(touchedElement[0])) {
        handleSelection(touchedElement);
      }
    }
  }

  function handleSelection(element) {
    if (gameOver) {
      return;
    }
    const isAlreadySelected = element[0].classList.contains('selected');

    if (isAlreadySelected) {
      removeFromSelection(element[0]);
      return;
    }

    if (isMoveLegal(element[0])) {
      if (currentSelection.length === 0) { // if "first" selection
        startingLetter = element[0];
        addToSelection(element[0]);
      } else if (currentSelection.length === 1) { // if "second" selection
        if (isAdjacentToFirstLetter(element[0])) {
          setDirection(element[0]);
          addToSelection(element[0]);
        } else {
          resetMove();
          startingLetter = element[0];
          addToSelection(element[0]);
        }
      } else {
        addToSelection(element[0]);
      }
    } else {
      resetMove();
      startingLetter = element[0];
      addToSelection(element[0]);
    }

    constructWord();
  }

  // UTILS

  function constructWord() {
    if (currentSelection.length > 0) {  // constructs word from selection
      let constructedWord = '';
      for (let i = 0; i < currentSelection.length; i++) {
        constructedWord += currentSelection[i].getAttribute('value');
      }
      checkWord(constructedWord);
    }
  }
  
  //ARIA
  var wordSearchAria = [];
  if( $("body").hasClass("latin") ) wordSearchAria = wordSearchLAT;
  if( $("body").hasClass("german") ) wordSearchAria = wordSearchDE1;
  if( $("body").hasClass("german osnovna") ) wordSearchAria = wordSearchDE5;
  if( $("body").hasClass("italian") ) wordSearchAria = wordSearchIT1;
  if( $("body").hasClass("italian5") ) wordSearchAria = wordSearchIT5;

  var page = window.location.pathname.split("/").pop();
  generateAria(page, gameId);
  
  function generateAria(page,gameId) {
	  		$.each(wordSearchAria, function( index, value ) {
			var findPage = value.page.toLowerCase().indexOf(page.toLowerCase()) >= 0;
			var findGame = value.id.toLowerCase() == gameId.toLowerCase();
			if (findPage && findGame){
				$(gameContainer).closest(".wordsearch-box").attr("aria-label", value.tekst);
			}
		});
  }
}

function invertString(string) {
  const wordArray = string.split('');
  const invertedArray = wordArray.reverse();
  const invertedString = invertedArray.join('');

  return invertedString;
}

// GRID GENERATOR

const LETTERS = 'abcdefghijklmnopqrstuvwxyz'; // letters used for filler
const MAXATTEMPTS = 20;                       // maximum amount of times to place a word

function generateGrid(words, options) {
  if (!words || !words.length) {
    console.log('Nisu pronadene rijeci');
    return false;
  }
  const width = options.width ? options.width : 15;
  const height = options.height ? options.height : 15;
  const letters = options.letters ? options.letters : LETTERS;

  // sort the words by length (biggest first)
  words.sort(function (a, b) {
    return a.length < b.length ? -1 : 1;
  });

  // populate the grid with empty arrays
  let grid = new Array(height);
  for (let i = 0; i < grid.length; i++) {
    grid[i] = new Array(width);
  }

  let unplaced = [];

  // loop the words
  for (let i = 0; i < words.length; i++) {
    const word = originalword = words[i];

    // pick a random spot
    // try to place the word in the grid
    let attempts = 0;
    while (attempts < MAXATTEMPTS) {
      // determine the direction
      const direction = Math.floor(Math.random() * 4);
      const info = directionInfo(word, direction, width, height);

      // word is too long
      if (info.maxx < 0 || info.maxy < 0 || info.maxy < info.miny || info.maxx < info.minx) {
        unplaced.push(originalword);
        break;
      }

      // random starting point
      let x = ox = Math.round(Math.random() * (info.maxx - info.minx) + info.minx);
      let y = oy = Math.round(Math.random() * (info.maxy - info.miny) + info.miny);

      // check for collisions
      let placeable = true;
      let count = 0;
      for (let l = 0; l < word.length; l++) {
        const charInGrid = grid[y][x];

        if (charInGrid) { // check if there is a character in the grid
          if (charInGrid !== word.charAt(l)) {
            // not the same letter
            placeable = false;
            break;
          } else {
            // same letter
            count++;
          }
        }
        y += info.dy;
        x += info.dx;
      }
      if (!placeable || count >= word.length) {
        attempts++;
        continue;
      }

      // the word is placeable
      x = ox;
      y = oy;
      for (let l = 0; l < word.length; l++) {
        grid[y][x] = word.charAt(l);

        y += info.dy;
        x += info.dx;
      }
      break;
    } // end placement while loop

    if (attempts >= 20) unplaced.push(originalword);
  } // end word loop

  // put in filler characters
  for (let i = 0; i < grid.length; i++) {
    for (let j = 0; j < grid[i].length; j++) {
      if (!grid[i][j]) {
        grid[i][j] = letters.charAt(
          Math.floor(Math.random() * letters.length)
        );
      }
    }
  }

  // some foreign chars are not supported (or their uppercase version), this is the place to customize or apply conditions
  const formattedLetters = grid.map(function (letterRow) {
    return letterRow.map(function (letter) {
      if (letter == 'ß' || letter == '\u00DF' || letter == '\u1E9E') {
        return letter.toLowerCase(); // in case its a uppercase variant
      } else {
        return letter.toUpperCase();
      }
    });
  });

  const formattedGrid = formattedLetters.map(function (row) { return row.join('') });

  if (unplaced.length) {
    console.log('Neke rijeci nisu stale u osmosmjerku, molimo pokusajte osvjeziti stranicu');
    console.log(unplaced);
  }

  return { grid: formattedGrid, unplaced: unplaced };
}

function directionInfo(word, direction, width, height) {
  // determine the bounds
  let minx = 0, miny = 0;
  let maxx = width - 1;
  let maxy = height - 1;
  let dx = 0, dy = 0;
  switch (direction) {
    case 0: // up-right
      maxy = height - 1;
      miny = word.length - 1;
      dy = -1;
      maxx = width - word.length;
      minx = 0;
      dx = 1;
      break;
    case 1: // right
      maxx = width - word.length;
      minx = 0;
      dx = 1;
      break;
    case 2: // down-right
      miny = 0;
      maxy = height - word.length;
      dy = 1;
      maxx = width - word.length;
      minx = 0;
      dx = 1;
      break;
    case 3: // down
      miny = 0;
      maxy = height - word.length;
      dy = 1;
      break;
    default:
      break;
  }
  return {
    maxx: maxx,
    maxy: maxy,
    minx: minx,
    miny: miny,
    dx: dx,
    dy: dy
  }
}